/*
 * read_in.c
 *
 * Ruby extension library to add IO#read_in(buf),
 * which read data into a fixed size string buf.
 */

#include "ruby.h"
#include "rubyio.h"
#include "rubysig.h"


/*
 * copied from ruby/io.c.
 */
#ifdef _STDIO_USES_IOSTREAM  /* GNU libc */
#  ifdef _IO_fpos_t
#    define READ_DATA_PENDING(fp) ((fp)->_IO_read_ptr != (fp)->_IO_read_end)
#  else
#    define READ_DATA_PENDING(fp) ((fp)->_gptr < (fp)->_egptr)
#  endif
#elif defined(FILE_COUNT)
#  define READ_DATA_PENDING(fp) ((fp)->FILE_COUNT > 0)
#elif defined(__BEOS__)
#  define ReadDataPending(fp) (fp->_state._eof == 0)
#elif defined(USE_CWGUSI)
#  define READ_DATA_PENDING(fp) (fp->state.eof == 0)
#else
/* requires systems own version of the ReadDataPending() */
extern int ReadDataPending();
#  define READ_DATA_PENDING(fp) ReadDataPending(fp)
#endif

#define READ_CHECK(fp) do {\
    if (!READ_DATA_PENDING(fp)) {\
        rb_thread_wait_fd(fileno(fp));\
        rb_io_check_closed(fptr);\
     }\
} while(0)

static size_t
io_fread(ptr, len, f)
    char *ptr;
    size_t len;
    FILE *f;
{
    size_t n = len;
    int c;

    while (n--) {
	if (!READ_DATA_PENDING(f)) {
	    rb_thread_wait_fd(fileno(f));
	}
	TRAP_BEG;
	c = getc(f);
	TRAP_END;
	if (c == EOF) {
	    *ptr = '\0';
	    break;
	}
	*ptr++ = c;
    }
    return len - n - 1;
}

static VALUE
rb_read_in(io, buf)
    VALUE io, buf;
{
    OpenFile *fptr;
    FILE *f;
    int n;

    if (RSTRING(buf)->len == 0)
    {
	rb_raise(rb_eRuntimeError, "buffer must not be an empty string.");
    }

    GetOpenFile(io, fptr);
    rb_io_check_readable(fptr);
    f = GetReadFile(fptr);

    READ_CHECK(f);
    n = io_fread(RSTRING(buf)->ptr, RSTRING(buf)->len, f);
    if (n == 0)
    {
	if (feof(f)) return Qnil;
	rb_sys_fail(fptr->path);
    }
    OBJ_TAINT(buf);

    return INT2FIX(n);
}

Init_read_in()
{
    rb_define_method(rb_cIO, "read_in", rb_read_in, 1);

    return;
}
